/*************************************************************************
 * File 	: kernel/ktau/ktau_inst.c
 * Version	: $Id: ktau_inst.c,v 1.1 2006/11/12 01:24:48 anataraj Exp $
 *
 ************************************************************************/

#include <linux/kernel.h>
#include <linux/config.h>		/* CONFIG_XXXX */
#include <linux/slab.h>			/* Kmalloc */	
#include <linux/sched.h>		/* task_struct def */
#include <linux/interrupt.h>
#include <linux/spinlock.h>        	/* spinlock */

#include <asm/atomic.h>			/* atomic */
#include <asm/current.h>		/* get_current */
#include <asm/system.h>			/* irqs_disabled */

/* For debugging message */
#define TAU_NAME "ktau_inst"
#define PFX TAU_NAME
//#define TAU_DEBUG 		1

/* KTAU header files */
#include <linux/ktau/ktau_print.h>
#include <linux/ktau/ktau_inst.h>
#include <linux/ktau/ktau_datatype.h>
#include <linux/ktau/ktau_hash.h>
#include <linux/ktau/ktau_perf.h>
#include <linux/ktau/ktau_bootopt.h>

#define KTAU_START_INDEX	399	/* Dynamic index start at 400 */

/*--------------------- GLOBAL INDEX  ---------------------*/

#include "ktau_linux-2.4.c"

/*
 * Global function index
 */
atomic_t ktau_global_index = ATOMIC_INIT(KTAU_START_INDEX);

/* 
 * This function is used to increment the global_ktau_index.
 */
/*
unsigned int get_ktau_index(void){
	return atomic_inc_return(&ktau_global_index);	
}
*/

/*
 * Increment global function index by 'incr'
 * then Return the global function index for hash table
 * incremented by 1 from original value (i.e same return
 * value as from get_kau_index).
 */
/*
unsigned int incr_ktau_index(int incr) {
        return atomic_add_return(incr, &ktau_global_index);
}
*/

void reset_ktau_index(void){
	atomic_set(&ktau_global_index,KTAU_START_INDEX);
}
/*--------------------------- TIMER ---------------------------------*/

#ifdef CONFIG_X86   

/*
 * The low-level Pentium-class timer    
 */
/*
static __inline__ unsigned long long int ktau_rdtsc(void)
{
     unsigned long long int x;
     __asm__ volatile (".byte 0x0f, 0x31" : "=A" (x));
     return x;
}
*/
#else /*CONFIG_X86*/
#ifdef CONFIG_PPC32
/*
 * The low-level PowerPC timer (TimeBase)   
 */
//static __inline__ unsigned long long ktau_rdtsc(void) {
unsigned long long ktau_rdtsc(void) {

        unsigned long long time = 0;
        unsigned long lword = 0, hword = 0;

        asm volatile (
                "rdtscppc: \n\t"
                "mfspr  %%r27, 269 \n\t"
                "mfspr  %%r28, 268 \n\t"
                "mfspr  %%r29, 269 \n\t"
                "cmpw   %%r29, %%r27 \n\t"
                "bne    rdtscppc \n\t"
                "ori    %0,%%r28,0 \n\t"
                "ori    %1,%%r29,0 \n\t"
                : "=r" (lword), "=r" (hword)
                :
                : "r27", "r28", "r29"
        );

        time = hword;

        return ( (time << 32) | lword );
}
#endif /*CONFIG_PPC32*/
#endif /*CONFIG_X86*/

/*---------------------VERIFICATION / DEBUGGING Functions ---------------------*/
/* 
 * This function dumps a callstack of this process from KTAU_TRACE 
 * to help debugging
 */ 
void dump_call_stack(int stackSize){
#ifdef CONFIG_KTAU_TRACE

#ifdef CONFIG_KTAU_BOOTOPT
	if(!ktau_verify_bootopt(KTAU_TRACE_MSK))
		return;
#endif /* CONFIG_KTAU_BOOTOPT*/

	int i;
	ktau_trace *c = current->ktau->cur_trace;
	ktau_trace *h = current->ktau->head_trace;
	ktau_trace *t = current->ktau->trbuf;
	//ktau_trace *l = current->ktau->last_dump;

	printk("KTAU_TRACE STACK:\n");
	for(i=0;i<=stackSize && c!=h ;i++){
		if(c > t)
			c--;
		else
			c = (t + KTAU_TRACE_LAST);
		
		printk("tsc  :%20llu, ",c->tsc);
		printk("addr :%15x, ",c->addr);
		printk("type :%5u\n",c->type);
	}
#endif /*CONFIG_KTAU_TRACE */
}

/*
 * This function verifies that the inclusive > exclusive time.
 */
int verify_excl_incl(unsigned long long excl, unsigned long long incl, 
		unsigned long long diff_time, h_ent* cur_ent)
{
		if(excl > incl){
			printk("verify_excl_incl: Exclusive time is greater than inclusive time.\nPID  = %d\nExcl = %llu\nIncl = %llu\n",
				current->pid,excl, incl);
		
			printk("in function = %x\n",cur_ent->addr);
			printk("stackDepth = %d\n",current->ktau->stackDepth);
			printk("nestingLevel = %d\n",cur_ent->nestingLevel);
			printk("diff_time = %llu\n",diff_time);
			dump_call_stack(24);
			return(-1);
		}
		return(0);
}

/*---------------------KTAU INSTRUMENTATION ---------------------*/
/*
 * This fuction is the start ONLY TRACE
 */
void _ktau_start_trace(unsigned int addr, unsigned int grp){
	unsigned long flags = 0;

	ktau_spin_lock(&current->ktau->hash_lock,flags);
	/********************************************************/
#ifdef CONFIG_KTAU_TRACE
#ifdef CONFIG_KTAU_BOOTOPT
	if(unlikely(ktau_verify_bootopt(KTAU_TRACE_MSK)))
#endif /* CONFIG_KTAU_BOOTOPT*/
		trace_buf_write(ktau_rdtsc(), addr, KTAU_TRACE_START);
#endif /*CONFIG_KTAU_TRACE*/
	/********************************************************/
	ktau_spin_unlock(&current->ktau->hash_lock,flags);
}

/*
 * This fuction is the stop ONLY TRACE
 */
void _ktau_stop_trace(unsigned int addr, unsigned int grp){
	unsigned long flags = 0;

	ktau_spin_lock(&current->ktau->hash_lock,flags);
	/********************************************************/
#ifdef CONFIG_KTAU_TRACE
#ifdef CONFIG_KTAU_BOOTOPT
	if(unlikely(ktau_verify_bootopt(KTAU_TRACE_MSK)))
#endif /* CONFIG_KTAU_BOOTOPT*/
		trace_buf_write(ktau_rdtsc(), addr, KTAU_TRACE_STOP);
#endif /*CONFIG_KTAU_TRACE*/
	/********************************************************/
	ktau_spin_unlock(&current->ktau->hash_lock,flags);
}


/*
 * This fuction is the start profile
 */
void _ktau_start_prof(unsigned int index,unsigned int addr, unsigned int grp){
	unsigned long flags = 0;

	//AN: Moved this statement below ONLY for Profiling.
	//h_ent *cur_ent = (h_ent*)(current->ktau->ktau_hash[index]);
	ktau_spin_lock(&current->ktau->hash_lock,flags);
	
	/********************************************************/
#ifdef CONFIG_KTAU_TRACE
#ifdef CONFIG_KTAU_BOOTOPT
	if(unlikely(ktau_verify_bootopt(KTAU_TRACE_MSK)))
#endif /* CONFIG_KTAU_BOOTOPT*/
		trace_buf_write(ktau_rdtsc(), addr, KTAU_TRACE_START);
#endif /*CONFIG_KTAU_TRACE*/
	/********************************************************/

#ifndef CONFIG_KTAU_TRACE
	//AN: Moved from top - only get this for profiles - not for traces
	h_ent *cur_ent = (h_ent*)(current->ktau->ktau_hash[index]);
	/*
	 * Update hash table
	 */
	if(unlikely(!(cur_ent))) {
		//AN_20060419_OPTIMZ
		//current->ktau->ktau_hash[index] = kmalloc(sizeof(h_ent),GFP_ATOMIC);
		if(!current->ktau->hash_mem) { 
			goto ktau_unlock;
		}
		current->ktau->ktau_hash[index] = (current->ktau->hash_mem+index);
		current->ktau->ent_count++;
		cur_ent = current->ktau->ktau_hash[index];
		cur_ent->type 			= KTAU_TIMER;
		cur_ent->data.timer.count 	= 0;
		cur_ent->data.timer.incl 	= 0;
		cur_ent->data.timer.excl 	= 0;
		cur_ent->addr 	 		= addr;
		cur_ent->nestingLevel		= 0;
	}

	/*********************************Simple Recursive START logic*******************************/
	/* In this version, we make every function to be reentrantable by keeping multiple
	 * timer.start in an array call reent. This array is used in all cases.
	 */

	// 0. Checking if it is exceeding the stackDepth.
	// AN_20060419_OPTIMZ
#ifdef CONFIG_KTAU_DEBUG
	if(current->ktau->stackDepth >= KTAU_NESTED_LEVEL){
		err("ktau_start_prof: Exceeding stackDepth allowed:\nPID:  %d, stackDepth: %u",
				current->pid,
				current->ktau->stackDepth);
		dump_call_stack(24);
		goto ktau_unlock;
	}
#endif /*CONFIG_KTAU_DEBUG*/

	// 1. get the start time
	if(likely(current->ktau->reent[current->ktau->stackDepth] == 0)) {
		current->ktau->reent[current->ktau->stackDepth] = ktau_rdtsc();

		info("ktau_start_prof: kernel function= %x, Level= %3u, PID = %d"
				,cur_ent->addr,cur_ent->nestingLevel,current->pid);
	}
	else {
		/* If the start time is not zero, there is something wrong with the logic. */	
		err("ktau_start_prof: ERROR in kernel profiling logic. PID=%d\naddr=%x\nindex=%d\nstackDepth=%d\nnestingLevel=%d\n",
				current->pid, addr, index, current->ktau->stackDepth, 
				cur_ent->nestingLevel);

		dump_call_stack(24);
	}

	// 2. Now increment stackDepth
	current->ktau->stackDepth++;

	/* 3. Increment this function's nestingLevel. This is to seperately track 
	 *    only this function's stack.
	 */
	cur_ent->nestingLevel++;

	/**************************************** END of Start Logic ******************************/

#endif /* ifndef CONFIG_KTAU_TRACE */

ktau_unlock:
	ktau_spin_unlock(&current->ktau->hash_lock,flags);
}

/*
 * This fuction is the wrapper for the _ktau_start_prof
 */
//void ktau_start_prof(unsigned int index,unsigned int addr,unsigned int grp){
//#ifdef CONFIG_KTAU_BOOTOPT
//	if(ktau_verify_bootopt(grp))
//#endif /* CONFIG_KTAU_BOOTOPT*/
//		_ktau_start_prof(index,addr, grp);		
//}

void _ktau_stop_prof(unsigned int index, unsigned int addr, unsigned int grp){
	unsigned long flags = 0;
	unsigned long long diff_time = 0, lc_excltime = 0;
	h_ent *cur_ent = NULL;

	ktau_spin_lock(&current->ktau->hash_lock,flags);

	/********************************************************/
	// read the STOP time
	diff_time = ktau_rdtsc();


#ifdef CONFIG_KTAU_TRACE
#ifdef CONFIG_KTAU_BOOTOPT
	if(unlikely(ktau_verify_bootopt(KTAU_TRACE_MSK)))
#endif /* CONFIG_KTAU_BOOTOPT*/
		trace_buf_write(diff_time, addr, KTAU_TRACE_STOP);
#endif /*CONFIG_KTAU_TRACE*/


#ifndef CONFIG_KTAU_TRACE
	cur_ent = (h_ent*)(current->ktau->ktau_hash[index]);
	if(unlikely(!(cur_ent))) {
		printk("ktau_stop_prof: Error. Stopping with no start. index = %u, addr = %x\n",
				index, addr);	
		goto out_unlock;
	}

	/***************************Simple Recursive Incl/Excl Time Logic******************/

	// 1. --- Decrement StackDepth
	current->ktau->stackDepth--;

	// 2. --- Decrement the nesting level
	cur_ent->nestingLevel--;

	// 3. --- Increment Call-count
	cur_ent->data.timer.count++;

	// 4. --- Calc incl time and store in "diff_time"
	// We calc-incl time using reent[] array. 
	diff_time = diff_time - current->ktau->reent[current->ktau->stackDepth];

	/* set that array[index] value to ZERO for future use. */
	current->ktau->reent[current->ktau->stackDepth] = 0;

	// 5. --- Add that incl time to Total incl time of function
	if(likely(cur_ent->nestingLevel == 0)) {
		/* Incl-time of re-ent/recursive functions must be ADDED only ONCE *
		 * i.e. the incl-time of the outer-most invocation of that func    *
		 * is used. The inc-times of other child-reent calls of the same   *
		 * function are included in this incl-time. */
		cur_ent->data.timer.incl += diff_time;
	}

	// 6. --- Help parent-excl time calc: add our incl to accum[] at current stackDepth
	current->ktau->accum[current->ktau->stackDepth] += diff_time; 
	
	// 7. --- Calc our excl time: our incl - accum-incl of children seen : AND add it to tot func. excl time
        lc_excltime = (diff_time - current->ktau->accum[current->ktau->stackDepth+1]);
        cur_ent->data.timer.excl += lc_excltime;

	// 8. --- After use, Reset accum-child incl time to ZERO, so that it can be used next time
	current->ktau->accum[current->ktau->stackDepth+1] = 0;

#ifdef CONFIG_KTAU_MERGE
        /* Update ktau_state at every level */
        if(unlikely(current->ktau->merge.uptr)) {
                if(current->ktau->stackDepth == 0) {
                        update_mergestate(&(current->ktau->merge), diff_time, diff_time, 1 /* incr by 1 */, 0 /*merge_kernel*/);
                }
                if(cur_ent->nestingLevel == 0) {
                        //then merge incl and excl. otherwise only excl time
                        update_mergestate(&(current->ktau->merge), diff_time, lc_excltime, 1 /*incr by 1*/, grp);
                } else {
                        update_mergestate(&(current->ktau->merge), 0, lc_excltime, 1 /*incr by 1*/, grp);
                }
        }
#endif /*CONFIG_KTAU_MERGE*/

	/* Eventually, we have to verify that inclusive time >= exclusive time */
#ifdef CONFIG_KTAU_DEBUG
	if(cur_ent->nestingLevel == 0){
		verify_excl_incl(cur_ent->data.timer.excl, cur_ent->data.timer.incl, diff_time, cur_ent); 
	}
#endif /*CONFIG_KTAU_DEBUG*/

#endif /* ifndef CONFIG_KTAU_TRACE */

out_unlock:
	ktau_spin_unlock(&current->ktau->hash_lock,flags);
}

/*
 * This fuction is the wrapper for the _ktau_stop_prof
 */
//void ktau_stop_prof(unsigned int index,unsigned int addr,unsigned int grp){
//#ifdef CONFIG_KTAU_BOOTOPT
//	if(ktau_verify_bootopt(grp))
//#endif /*CONFIG_KTAU_BOOTOPT*/
//		_ktau_stop_prof(index,addr, grp);		
//}

/*
 * This fuction is the event profile
 */
void ktau_event_prof(unsigned int index, unsigned int addr){
	/*
	unsigned long flags = 0;
	h_ent *cur_ent = (h_ent*)(current->ktau->ktau_hash[index]);

	ktau_spin_lock(&current->ktau->hash_lock,flags);

	if(!(cur_ent)){
		current->ktau->ktau_hash[index] = kmalloc(sizeof(h_ent),GFP_ATOMIC);
		current->ktau->ent_count++;
		cur_ent = current->ktau->ktau_hash[index];
		cur_ent->type = KTAU_EVENT;
		cur_ent->data.timer.count = 0;
		cur_ent->data.timer.incl = 0;
		//cur_ent->data.timer.addr  = addr;
	}
	cur_ent->data.timer.count++;

	ktau_spin_unlock(&current->ktau->hash_lock,flags);
	*/
}
