#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>

#include "ktau_proc_interface.h"
#include "ktau_proc_map.h"
#include "ktau_and_tau.h"
#include "ktau_diff_profiles.h"

/* Input: filename
   Output: stdout
*/

void usage(char* cmd) {
	printf("prompt>%s <binary-trace-filename> > <ascii-trace-filename>\n\n", cmd);
}

int main(int argc, char* argv[]) {
	int ret = 0, fd = 0, noprofs = 0;
	struct stat statbuf;
	char* filebuf = NULL, *currptr = NULL;
	ktau_output* output = NULL;

	if(argc != 2) {
		usage(argv[0]);
		exit(-1);
	}

	ret = stat(argv[1], &statbuf); 
	if(ret) {
		perror("Cannot stat input file");
		usage(argv[0]);
		return ret;
	}

	filebuf = (char*)malloc(statbuf.st_size);
	if(!filebuf) {
		perror("malloc failed");
		exit(-1);
	}
	
	fd = open(argv[1], O_RDONLY);
	if(fd < 0) {
		perror("open failed");
		usage(argv[0]);
		goto free_out;
	}

	currptr = filebuf;
	while((ret = read(fd, currptr, statbuf.st_size)) > 0) {
		currptr += ret;
	}
	if(ret < 0) {
		perror("read failed");
		goto free_out;
	}

        noprofs = unpack_bindata(KTAU_TRACE, filebuf, statbuf.st_size, &output);
        if(!noprofs || !output) {
                printf("ktau_output from unpack_bindata is:%x. no-profs is:%d\n", output, noprofs);
		goto free_out;
        }

	print_many_profiles(KTAU_TRACE, stdout, output, noprofs);

free_out:
	free(filebuf);

	return 0;
}


