/*************************************************************
 * File         : user-src/src/runktau/ktauresize.c
 * Version      : $Id: ktauresize.c,v 1.1 2006/11/12 01:24:52 anataraj Exp $
 *
 ************************************************************/

/************************************************
 * KTAU Performance Analysis Suite		*
 * 						*
 * ktauresize:					*
 *	- used to resize kernel data-structs	*
 *	- currently only to resize TRACE buf	*
 *	- to avoid loss of trace data		*	
 * 						*
 * Usage: ktauresize -s <size of trace buffer> 	*
 *		-n <no of pids> -p <pid list>	*
 * 						*
 ***********************************************/

#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/utsname.h>

#include "ktau_proc_interface.h"
#include "ktau_proc_map.h"
#include "ktau_and_tau.h"
#include "ktau_diff_profiles.h"

//#include "../ktaud/ktau_timer.c"

void usage(char* cmd) {
	fprintf(stderr, "USAGE:\n");
	fprintf(stderr, "prompt>%s <SIZE of tracebuf> <no of processes to resize> <pid list...>\n\n", cmd);
	fprintf(stderr, "For help:\n");
	fprintf(stderr, "prompt>%s -h\n\n", cmd);
	fprintf(stderr, "EXAMPLE 1: Resize Trace buffer of task-pid:777 to 10000 entries.\n");
	fprintf(stderr, "prompt>%s 10000 1 777\n\n", cmd);
	fprintf(stderr, "EXAMPLE 2: Resize Trace buffer of ALL tasks(pid:0) to 10000 entries.\n");
	fprintf(stderr, "prompt>%s 10000 0\n\n", cmd);
	fflush(stderr);
}

#define MAX_NO_PIDS 128

int main(int argc, char* argv[])
{
	int nopids = 0, resize = 0;
	int pids[MAX_NO_PIDS];

	if( (argc < 3) || ((argc > 1) && (!strcmp(argv[1],"-h")))) {
		usage(argv[0]); 
		exit(-1);
	}

	resize = atoi(argv[1]);
	if(resize <= 0) {
		usage(argv[0]); 
		exit(-1);
	}

	nopids = atoi(argv[2]);
	if(nopids < 0) {
		usage(argv[0]); 
		exit(-1);
	}

	if(nopids > 0) {
		int i = 0;
		for(; i<nopids; i++) {
			pids[i] = atoi(argv[3+i]);
			if(pids[i] <= 0) {
				usage(argv[0]); 
				exit(-1);
			}
		}
	}

	int* pidlist = NULL;
	if(nopids > 0) pidlist = pids;
	
	printf("Resize ret: %d\n", ktau_trace_resize(resize, KTAU_TRACE, 0, pidlist, nopids, 0, NULL));

	return 0;
}


